//
// Copyright (c) 2001, Adam Dunkels.
// Copyright (c) 2005, Invector Embedded Technologies
// Copyright (c) 2006, Olimex Ltd
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
// 3. The name of the author may not be used to endorse or promote
//    products derived from this software without specific prior
//    written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
// OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
// GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// ---------------------------------------------------------------------------
//
// Module:			DM9000 ethernet routines for 8051 and SDCC port of Adam Dunkels'
//              uIP v0.9
// Filename:		DM9000.c
// Version:			0.2
// Date:			  30/MAR/2005
// Author:			Adam Dunkels, A.Pemberton, Pontus Oldberg, Ivan Vasilev
// Function:		DM9000 and other H/W definitions
//
// Change History:	
//              2006-03-15 PO Cleaned up code
//              2005-09-14 AP Basic DM9000 module - frills removed
//			  2006-07-17 IV Adapted to use LPC's IO functions, added support for
//								  automatic detection of databus width
// ---------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dm9000.h"		  // Physical Layer
#include <iolpc2114.h>
#define IOWAIT      IO0PIN_bit.P0_19
#define IOWAIT_DIR  IO0DIR_bit.P0_19


//#define VERSION 1_1

//---------------------------------------------------------------------------------
// Static variables for DM9000
//---------------------------------------------------------------------------------

extern int TX_EventPending;	// the DM9000 hardware receive event
extern int ARP_EventPending;	// trigger the arp timer event

u8_t* PORT_DAT = (u8_t *)0x81000004;
u8_t* PORT_REG = (u8_t *)0x81000000;
u32_t* PORT_DAT32 = (u32_t *)0x81000004;
u32_t* PORT_REG32 = (u32_t *)0x81000000;
u16_t* PORT_DAT16 = (u16_t *)0x81000004;
u16_t* PORT_REG16 = (u16_t *)0x81000000;


//---------------------------------------------------------------------------------
// Local prototypes
//---------------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Write Command to NIC
//-----------------------------------------------------------------------------


void write_nicreg(u8_t IOaddr, u8_t val)
{
  while (!IOWAIT);
  *PORT_REG = (u8_t)IOaddr;
  *PORT_DAT = (u8_t)val;

}

//-----------------------------------------------------------------------------
// Write Data to NIC
//-----------------------------------------------------------------------------
void write_nicdata(u8_t val)
{
  //NIC_DAT = val;
  *PORT_DAT = (u8_t)val;
}

void write_nicdata32(u32_t val)
{
  //NIC_DAT = val;
  *PORT_DAT32 = (u32_t)val;
}

void write_nicdata16(u32_t val)
{
  //NIC_DAT = val;
  *PORT_DAT16 = (u16_t)val;
}

//-----------------------------------------------------------------------------
// Write Command or Data to already selected register
//-----------------------------------------------------------------------------
void set_nicreg(u8_t IOaddr)
{
  //NIC_REG = IOaddr;
  *PORT_REG = (u8_t)IOaddr;
}

//-----------------------------------------------------------------------------
// Write Command to PHY
//-----------------------------------------------------------------------------

void PHY_write(u8_t IOaddr, u16_t val)
{
 	write_nicreg(DM9000_EPCR,0x0a);					// Set EEPROM & PHY Control Reg to write
 	write_nicreg(DM9000_EPAR,((IOaddr & 0x3f) | DM9000_PHY));	// Set PHY Address Reg to PHY Address
  write_nicreg(DM9000_EPDRH,(u8_t)(val >> 8));	// write high_byte of PHY data
  write_nicreg(DM9000_EPDRL,(u8_t)(val));			// write low_byte of PHY data
 	write_nicreg(DM9000_EPCR,0x00);					// Clear EEPCR
}

//--------------------------------------------------------------------------------
// Read Status from NIC
//--------------------------------------------------------------------------------
u8_t read_nicreg(u8_t IOaddr)
{

  *PORT_REG = (u8_t)IOaddr;
  return (*PORT_DAT);
}


//--------------------------------------------------------------------------------
// Read Data from NIC from preselected address
//--------------------------------------------------------------------------------
u32_t read_nicdata32(void)
{
  return (*PORT_DAT32);
}

u16_t read_nicdata16(void)
{
  return (*PORT_DAT16);
}

u8_t read_nicdata(void)
{
  return (*PORT_DAT);
}

//-----------------------------------------------------------------------------
// Non-interrupt delay approx 1mS to 255mS
//-----------------------------------------------------------------------------
void _wait_ms(u16_t count)
{
	u16_t i;

  for ( ; count > 0; count--) {
		for ( i = 10000; i > 0; i--);
	}
}

//-----------------------------------------------------------------------------
// u8_t InitDM9000(void);
//
// Init the PHY. If OK, return true
//-----------------------------------------------------------------------------
// Driver Initializing Steps

u8_t InitDM9000(void)
{

  //IO0DIR_bit.P0_15 = 1; //WAKEUP
  //IO0SET_bit.P0_15 = 1;
  IO0DIR_bit.P0_25 = 1; //RST
  IO0DIR_bit.P0_24 = 1; //PWR_RST
  IO0SET_bit.P0_25 = 1; //Active HIGH reset
  IO0CLR_bit.P0_24 = 1; //Acive LOW reset
  IO0DIR_bit.P0_19 = 0; //IOWAIT as input


  _wait_ms(50);
  IO0SET_bit.P0_24 = 1;
  _wait_ms(50);
  IO0CLR_bit.P0_25 = 1;
  _wait_ms(100);
  IO0DIR_bit.P0_15 = 0;
  _wait_ms(100);


  //1. If the internal PHY is required, the following steps are to active the internal PHY:
  // The default status of the DM9000 is to power down the internal PHY by setting the GPIO0.
  // Since the internal PHY have been powered down, the wakeup procedure will be needed to
  // enable the DM9000.
  write_nicreg(DM9000_GPCR, 0x07);    // Power PHY set the GPR to output
  write_nicreg(DM9000_GPR,  0x06);	  // Clear the GPIO lines (Led's off)
  //_wait_ms(1);

  //2. Program NCR register. Choose normal mode by setting NCR (reg_00h) bit[2:1] = 00h. The
  // system designer can choose the network operation such as setting internal/external PHY,
  // enable wakeup event or choose the full-duplex mode. Please refer to the DM9000 datasheet
  // ch.6.1 about NCR register setting.
  // Reset DM9000 (twice)
  write_nicreg(DM9000_NCR, 0x03);	// Reset
  //_wait_ms(1);
  write_nicreg(DM9000_NCR, 0x00);	// Normal mode
  //_wait_ms(1);

  write_nicreg(DM9000_NCR, 0x03);
  //_wait_ms(1);
  write_nicreg(DM9000_NCR, 0x00);
  //_wait_ms(1);

  //3. Clear TX status by reading NSR register (reg_01h). Bit[2:3] and bit[5] will be automatically
  // cleared by reading or writing 1. Please refer to the DM9000 datasheet ch.6.2 about NSR
  // register setting.
  read_nicreg(DM9000_NSR);
  //_wait_ms(1);

  //4. Read EEPROM (No EEPROM available - use scratchpad instead) data if
  // valid or required. By default, the array will have hard MAC loaded.

  //5. Set Node address 6 bytes from in physical address register (reg_10h~15h).
 	write_nicreg(DM9000_MAC_REG,  uip_ethaddr.addr[0]);
 	write_nicreg(DM9000_MAC_REG+1,uip_ethaddr.addr[1]);
 	write_nicreg(DM9000_MAC_REG+2,uip_ethaddr.addr[2]);
 	write_nicreg(DM9000_MAC_REG+3,uip_ethaddr.addr[3]);
 	write_nicreg(DM9000_MAC_REG+4,uip_ethaddr.addr[4]);
 	write_nicreg(DM9000_MAC_REG+5,uip_ethaddr.addr[5]);
        //_wait_ms(1);

  //6. Set Hash table 8 bytes from multicast address register (reg_16h~1Dh).
 	write_nicreg(DM9000_MLC_REG,  0xff);
 	write_nicreg(DM9000_MLC_REG+1,0xff);
 	write_nicreg(DM9000_MLC_REG+2,0xff);
 	write_nicreg(DM9000_MLC_REG+3,0xff);
 	write_nicreg(DM9000_MLC_REG+4,0xff);
 	write_nicreg(DM9000_MLC_REG+5,0xff);
 	write_nicreg(DM9000_MLC_REG+6,0xff);
 	write_nicreg(DM9000_MLC_REG+7,0xff);
        //_wait_ms(1);

  //7. reset Internal PHY if desired
  //	PHY_write(DM9000_BMCR, 0x8000);	// PHY reset
  //	 _wait_ms(1000);
  //	PHY_write(DM9000_BMCR, 0x1200);	// Auto negotiation
  //	 _wait_ms(1000);


  //8. Set IMR register (reg_FFh) bit[7]=1 to enable the SRAM read/write pointer which is the
  // automatic return function of the memory R/W address pointer; also set the receive int mask.
 	write_nicreg(DM9000_IMR, 0x81);
        //_wait_ms(1);

  //9. Depend on OS and DDK of the system to handle NIC interrupts.
  // Not applicable since the DM9000 is polled

  //10. Program IMR register (reg_FFh) bit[1:0] to enable the TX/RX interrupt. Before doing this,
  // the system designer needs to register the interrupt handler routine. For example, if the driver
  // needs to generate the interrupt after a package is transmitted, the interrupt mask register IMR
  // bit[1]=1 will be set. If the interrupt is generated by the DM9000 after receiving a packet, IMR
  // bit[0] should be set to 1.
  // Not applicable since the DM9000 is polled

  //11. Program RXCR register to enable RX. The RX function is enabled by setting the RX control
  // register (reg_05h) bit[0]=1. The choice of the other bits bit[6:0] depends on the system design.
  // Please refer to the DM9000 datasheet ch.6.6 about RXCR register setting.
 	write_nicreg(DM9000_RXCR, 0x31);	// Only MAC-packets and RX enable - no multicast packets
        //_wait_ms(1);


  //12. NIC is being activated now.

      //  while (1) write_nicreg(DM9000_MAC_REG,  uip_ethaddr.addr[0]);


  //13. Check Vendor ID is Davicom (0x0A46)

  if ((read_nicreg(DM9000_VID) + (read_nicreg(DM9000_VID+1) << 8)) != DM9000_VendID)
    return TRUE;		// No chip found

  if ((read_nicreg(DM9000_PID) + (read_nicreg(DM9000_PID+1) << 8)) != DM9000_ProdID)
    return TRUE;		// No chip found

//  RX_length = 0;	

  return FALSE;
}


void DM9000_transmit_low(char buftmp[UIP_BUFSIZE], unsigned int TXlen)
{
	u8_t io_mode;
	u16_t i;
	u16_t tmplen;
	
	while (read_nicreg(DM9000_TCR) & 0x01); // wait transmit complete
	write_nicreg(DM9000_ISR, 0x02);

	io_mode = read_nicreg(DM9000_ISR)>>6;	// bits 1:0 of ISR(0xFE)

	set_nicreg(DM9000_MWCMD);
	
	if (io_mode == DM9000_BYTE_MODE)
	{
		for (i=0;i<TXlen;i++)
                {
                  while (!IOWAIT);
                  write_nicdata(buftmp[i]);
                }
	}
        else if (io_mode == DM9000_WORD_MODE)
        {
            tmplen = (TXlen+1)/2;

            for (i=0;i<tmplen;i++)
            {
              u16_t sendtmp;
              u16_t sendtmp1;
              u16_t sendtmp2;

              // HTONS
              sendtmp1 = (u16_t)(buftmp[1+i*2])<<8;
              sendtmp2 = (u16_t)(buftmp[i*2]);
              sendtmp = sendtmp1|sendtmp2;
              write_nicdata16(sendtmp);
            }
        }
        else if (io_mode == DM9000_DWORD_MODE)
	{
		tmplen = (TXlen+3)/4;

		for (i=0;i<tmplen;i++)
		{
			u32_t sendtmp;
			u32_t sendtmp1;
			u32_t sendtmp2;
			u32_t sendtmp3;
			u32_t sendtmp4;

			// Home-grown HTONL implementation
			sendtmp1 = (u32_t)(buftmp[3+i*4])<<24;
			sendtmp2 = (u32_t)(buftmp[2+i*4])<<16;
			sendtmp3 = (u32_t)(buftmp[1+i*4])<<8;
			sendtmp4 = (u32_t)(buftmp[i*4]);
			sendtmp = sendtmp1 | sendtmp2 | sendtmp3 | sendtmp4;
			
			write_nicdata32(sendtmp);
		}
	}
}


void DM9000_transmit(void) // horrendeously stupid implementation :(
{

  static u16_t i;
  static u16_t j;
  u8_t volatile *ptr; // general pointer for data
  char tmpbuf[UIP_BUFSIZE];

  ptr = &uip_buf[0];

  if (uip_buf[UIP_LLH_LEN + 9] == UIP_PROTO_TCP)
  {
    j = (UIP_LLH_LEN + UIP_TCPIP_HLEN + (((uip_buf[UIP_LLH_LEN + 32] >> 4) - 5) << 2));
  }
  else if (uip_buf[UIP_LLH_LEN + 9] == UIP_PROTO_ICMP)
  {
    j = (UIP_LLH_LEN + UIP_TCPIP_HLEN);
  }
  else if (uip_buf[UIP_LLH_LEN + 9] == UIP_PROTO_UDP)
  {
    j = (UIP_LLH_LEN + UIP_TCPIP_HLEN);
  }
  else
    j = (UIP_LLH_LEN + UIP_TCPIP_HLEN);

  for (i = 0; i < j; i++)
  {
    tmpbuf[i] = *ptr++;
  }

  if (i < uip_len)
  {
    for (ptr = uip_appdata; (i < uip_len) ; i++)
    {
			tmpbuf[i] = *ptr++;
    }
  }

	DM9000_transmit_low(tmpbuf, uip_len);
	write_nicreg(DM9000_TXPLH, (u8_t)(uip_len>>8));
	write_nicreg(DM9000_TXPLL, (u8_t)(uip_len));

	
	write_nicreg(DM9000_TCR, 1);		// start the transmission
}


u8_t DM9000_check_ready()
{
	return read_nicreg(DM9000_MRCMDX);
}

u16_t DM9000_receive(void)
{
	u8_t RX_ready;
	u16_t RX_status;
	u16_t RX_lenght;
	u8_t io_mode;
	u32_t len_tmp;
	u16_t i;
	
	do
	{
		RX_ready = DM9000_check_ready();
		if (RX_ready == 0x01)
		{
			io_mode = read_nicreg(DM9000_ISR)>>6;
			set_nicreg(DM9000_MRCMD);	// set read autoincrement
			
			if ( io_mode == DM9000_BYTE_MODE)
			{
				RX_status = (read_nicdata()+(read_nicdata()<<8));
				RX_lenght = (read_nicdata()+(read_nicdata()<<8));
				for (i=0;i<RX_lenght;i++)
				{
					uip_buf[i] = read_nicdata();
				}
				write_nicreg(DM9000_ISR, DM9000_RX_INTR);
				return RX_lenght;
			}
      if (io_mode == DM9000_WORD_MODE)
      {
        RX_status = read_nicdata16() & 0xff00;
        RX_lenght = read_nicdata16();
        len_tmp = (RX_lenght+1)/2;
        for (i=0;i<len_tmp;i++)
        {
          u16_t rcvtmp = read_nicdata16();
          uip_buf[i*2] = (u8_t)(rcvtmp & 0x00ff);
					uip_buf[i*2+1] = (u8_t)((rcvtmp>>8) & 0x00ff);
        }
        write_nicreg(DM9000_ISR, DM9000_RX_INTR);
        return RX_lenght;	
      }

			if ( io_mode == DM9000_DWORD_MODE)
			{
				len_tmp = read_nicdata32();
				RX_status = (u16_t)(len_tmp & 0x0ff00);
				RX_lenght = (u16_t)( (len_tmp>>16)&0xffff);                                                          				
				len_tmp = (RX_lenght+3)/4;
				for (i=0;i<len_tmp;i++)
				{
					u32_t rcvtmp = read_nicdata32();
					uip_buf[i*4] = (u8_t)(rcvtmp & 0x000000ff);
					uip_buf[i*4+1] = (u8_t)((rcvtmp>>8) & 0x000000ff);
					uip_buf[i*4+2] = (u8_t)((rcvtmp>>16) & 0x000000ff);
					uip_buf[i*4+3] = (u8_t)((rcvtmp>>24) & 0x000000ff);
				}
				write_nicreg(DM9000_ISR, DM9000_RX_INTR);
				return RX_lenght;				
			}
		}
		else if (RX_ready != 0x00)
		{
			write_nicreg(DM9000_IMR, 0x80);
			write_nicreg(DM9000_ISR, 0x0f);
			write_nicreg(DM9000_RXCR, 0x00);
			
			InitDM9000();
			return 0;			
		}
		
	} while (!(TX_EventPending || ARP_EventPending));
	return 0;
}

void eeprom_write(u8_t adress, u16_t data)
{
  write_nicreg(DM9000_EPAR, adress);
  write_nicreg(DM9000_EPDRL, (u8_t)(data&0xff));
  write_nicreg(DM9000_EPDRH, ((u8_t)((data>>8)&0xff)));

  write_nicreg(DM9000_EPCR, 0x12);

  _wait_ms(100);
   write_nicreg(DM9000_EPCR, 0x00);
}

u16_t eeprom_read(u8_t adress)
{
  u16_t res=0;

  write_nicreg(DM9000_EPAR, adress);
  write_nicreg(DM9000_EPCR, 0x04);

  while ( read_nicreg(DM9000_EPCR)&0x01);

  _wait_ms(10);
  write_nicreg(DM9000_EPCR, 0x00);

  res = (read_nicreg(DM9000_EPDRH)<<8)|(read_nicreg(DM9000_EPDRL));
  return res;
}
